-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
--    Copyright (c) 2002  Meindert Kuipers, Netherlands               --
--    meindert@kuipers.to                www.kuipers.to               --
--                                                  --
-- This program is free software; you can redistribute it and/or          --
-- modify it under the terms of the GNU General Public License            --
-- as published by the Free Software Foundation; either version 2         --
-- of the License, or (at your option) any later version.             --
--                                                  --
-- This program is distributed in the hope that it will be useful,        --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of         --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          --
-- GNU General Public License for more details.                     --
--                                                  --
-- You should have received a copy of the GNU General Public License        --
-- along with this program; if not, write to the Free Software            --
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA--
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

-- This version of the testbench is for the SmartRom
-- Code is cleaned up
-- Changed to 16-bits data and 20-bits addressing
-- Note that not the fullspace of the memories is used!

LIBRARY ieee;

USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;
USE ieee.std_logic_unsigned.all;
USE ieee.std_logic_arith.all;
USE std.textio.all;

ENTITY testbench IS
END testbench;

ARCHITECTURE behavior OF testbench IS

-- HP41 timing and interface constants
-- CLOCK and Cycle

constant ACCEL        : integer := 5;          -- acceleration factor
constant CLK_PERIOD   : time := 1000 ns/ACCEL; -- 1 us clock period of HP41
                                               -- shorter here for of simulation
constant CLK_HIGH     : time := CLK_PERIOD/10; -- define duty cycle, is here 10%
constant CLK_LAG      : time := CLK_PERIOD/2;  -- Lag of CLK01 behind CLK02
constant CLK_LENGTH   : integer := 55;         -- length of cycle - 1


-- ISA ADDRESS and INSTRUCTION
constant ISA_AD_START : integer := 16;      -- start of ADDRESS on ISA
constant ISA_AD_END   : integer := 31;      -- end of ADDRESS on ISA
constant ISA_IS_START : integer := 46;      -- start of INSTRUCTION on ISA
constant ISA_IS_END   : integer := 55;      -- end of INSTRUCTION on ISA

-- SYNC
constant SYNC_START   : integer := 46;      -- start of SYNC
constant SYNC_END     : integer := 55;      -- end of SYNC

-- ADDRESS and INSTRUCTION for WROM
constant DATA_AD_START  : integer := 14;      -- start of ADDRESS on DATA
constant DATA_AD_END    : integer := 29;      -- end of ADDRESS on DATA
constant DATA_IS_START  : integer := 2;       -- start of INSTRUCTION on DATA
constant DATA_IS_END    : integer := 11;      -- end of INSTRUCTION on DATA

-- Constants for EPROM and Stimulus files

constant addr_0k: natural := 0;
constant addr_2k: natural := 2047;             -- 2K EPROM for U2
constant addr_8k: natural := 8195;             -- 8K EPROM for L8
constant addr_64k:natural := 65536;            -- 64K FLASH/SRAM for SmartROM
                          -- we do not use the full 4 MBit in the simulator
constant SRAM_access    : time := 90 ns / ACCEL;     -- SRAM access time
constant FLASH_access   : time := 90 ns / ACCEL;     -- FLASH access time

-- Constants for External I/O Interface
constant X_WR_active    : time := 400 ns / ACCEL;    -- active time for X_WR
constant X_RD_active    : time := 400 ns / ACCEL;    -- active time for X_RD
constant X_delay        : time := 1000 ns / ACCEL;   -- delay after changing an I/O

-- Data types
subtype nybble   is std_logic_vector( 3 downto 0);   -- 4-bit nybble
subtype byte     is std_logic_vector( 7 downto 0);   -- regular byte
subtype tyte     is std_logic_vector( 9 downto 0);   -- HP41 10-bit word
subtype word     is std_logic_vector(15 downto 0);   -- regular 16-bit word
subtype xaddress is std_logic_vector(19 downto 0);   -- 20-bit address
subtype lword    is std_logic_vector(31 downto 0);   -- 32-bit longword for address
subtype data_reg is std_logic_vector(55 downto 0);   -- 56-bit HP41 data register

type data_char is array (natural range 0 to 13) of character;  -- for reading DATA register

-- memory types for simulation of SRAM and FLASH
type Flash_memory is array (natural range addr_0k to addr_64k) of word;
type SRAM_memory  is array (natural range addr_0k to addr_64k) of word;


-- HP41 Stimulus types and array
type hp41stim is record
  stim_Addr   : word;
  stim_Instr  : tyte;
  stim_SYNC   : std_logic;
  stim_DATA   : data_reg;
end record;

-- I/O Stimulus type
type IOstim is record
   stim_Mode   : integer;
   stim_Cmd    : nybble;
   stim_Data   : tyte;
   stim_Addr   : xaddress;
end record;


constant max_stim: natural := 500;
type stim_array is array(natural range 0 to max_stim) of hp41stim;


function hex2bin (h1: character) return natural is
  variable res: integer := 0;
begin
  case h1 is
    when '0' => res :=  0;
    when '1' => res :=  1;
    when '2' => res :=  2;
    when '3' => res :=  3;
    when '4' => res :=  4;
    when '5' => res :=  5;
    when '6' => res :=  6;
    when '7' => res :=  7;
    when '8' => res :=  8;
    when '9' => res :=  9;
    when 'A' => res := 10;
    when 'B' => res := 11;
    when 'C' => res := 12;
    when 'D' => res := 13;
    when 'E' => res := 14;
    when 'F' => res := 15;
    when others =>
  end case;
  return res;
end hex2bin;

COMPONENT SmartROM
  PORT(
    SYNC      : IN  std_logic;
    CLK01     : IN  std_logic;
    CLK02     : IN  std_logic;
    DATA      : IN  std_logic;
    ISA       : IN  std_logic;
    PWO       : IN  std_logic;

    ISA_Out   : OUT std_logic;
    ISA_OE    : OUT std_logic;
    DATA_Out  : OUT std_logic;
    DATA_OE   : OUT std_logic;
    SYNC_Out  : OUT std_logic;
    SYNC_OE   : OUT std_logic;
    FI_Out    : OUT std_logic;
    FI_OE     : OUT std_logic;

    XA        : OUT std_logic_vector(19 downto 00);
    XD        : INOUT std_logic_vector(09 downto 00);
    XXD       : IN std_logic_vector(15 downto 10);

    FLASH_CE  : OUT std_logic;
    FLASH_OE  : OUT std_logic;
    FLASH_WE  : OUT std_logic;
    SRAM_CE1  : OUT std_logic;
    SRAM_CE2  : OUT std_logic;
    SRAM_OE   : OUT std_logic;
    SRAM_WE   : OUT std_logic;
    MEM_RESET : OUT std_logic;
    RY_BY     : IN  std_logic;

    MLDL_SR   : IN  std_logic;
    SR_SET    : IN  std_logic_vector(1 downto 0);
    MLDL_DIS  : IN  std_logic;

    X_EN      : IN  std_logic;
    X_CLK     : IN  std_logic;
    X_STROBE  : IN  std_logic;
    X_DIN     : IN  std_logic;
    X_DOUT    : OUT std_logic;
    X_DAV     : OUT std_logic;
    X_BSY     : OUT std_logic;

    DSR_clk   : IN  std_logic;
    OSR_clk   : IN  std_logic;
    DSR_clk_out : OUT std_logic;
    OSR_clk_out : OUT std_logic

  );
END COMPONENT;


  SIGNAL SYNC     : std_logic;
  SIGNAL CLK01    : std_logic;
  SIGNAL CLK02    : std_logic;
  SIGNAL DATA     : std_logic;
  SIGNAL ISA      : std_logic;
  SIGNAL PWO      : std_logic;

  SIGNAL ISA_Out  : std_logic;
  SIGNAL ISA_OE   : std_logic;
  SIGNAL DATA_Out : std_logic;
  SIGNAL DATA_OE  : std_logic;
  SIGNAL SYNC_Out : std_logic;
  SIGNAL SYNC_OE  : std_logic;
  SIGNAL FI_Out   : std_logic;
  SIGNAL FI_OE    : std_logic;

  SIGNAL XA       : std_logic_vector(19 downto 00);
  SIGNAL XD       : std_logic_vector(09 downto 00);
 SIGNAL XXD      : std_logic_vector(15 downto 10);

  SIGNAL FLASH_CE : std_logic;
  SIGNAL FLASH_OE : std_logic;
  SIGNAL FLASH_WE : std_logic;
  SIGNAL SRAM_CE1 : std_logic;
  SIGNAL SRAM_CE2 : std_logic;
  SIGNAL SRAM_OE  : std_logic;
  SIGNAL SRAM_WE  : std_logic;
  SIGNAL MEM_RESET: std_logic;
  SIGNAL RY_BY    : std_logic;

  SIGNAL MLDL_SR  : std_logic;
  SIGNAL SR_SET   : std_logic_vector(1 downto 0);
  SIGNAL MLDL_DIS : std_logic;

  SIGNAL X_EN     : std_logic;
  SIGNAL X_CLK    : std_logic;
  SIGNAL X_STROBE : std_logic;
  SIGNAL X_DIN    : std_logic;
  SIGNAL X_DOUT   : std_logic;
  SIGNAL X_DAV    : std_logic;
  SIGNAL X_BSY    : std_logic;

  SIGNAL DSR_clk    : std_logic;
  SIGNAL OSR_clk    : std_logic;
  SIGNAL DSR_clk_out: std_logic;
  SIGNAL OSR_clk_out: std_logic;

  -- HP41 MLDL Testbench specific signals
  signal phase_counter : integer := 0 ;
  signal cycle_counter : integer := 0 ;

BEGIN

  uut: SmartROM PORT MAP(
    SYNC      => SYNC,
    CLK01     => CLK01,
    CLK02     => CLK02,
    DATA      => DATA,
    ISA       => ISA,
    PWO       => PWO,

    ISA_out   => ISA_out,
    ISA_OE    => ISA_OE,
    DATA_out  => DATA_out,
    DATA_OE   => DATA_OE,
    SYNC_out  => SYNC_out,
    SYNC_OE   => SYNC_OE,
    FI_out    => DATA_out,
    FI_OE     => DATA_OE,

    XA        => XA,
    XD        => XD,
    XXD       => XXD,

    FLASH_CE  => FLASH_CE,
    FLASH_OE  => FLASH_OE,
    FLASH_WE  => FLASH_WE,
    SRAM_CE1  => SRAM_CE1,
    SRAM_CE2  => SRAM_CE2,
    SRAM_OE   => SRAM_OE,
    SRAM_WE   => SRAM_WE,
    MEM_RESET => MEM_RESET,
    RY_BY     => RY_BY,

    MLDL_SR   => MLDL_SR,
    SR_SET    => SR_SET,
    MLDL_DIS  => MLDL_DIS,

    X_EN      =>  X_EN,
    X_CLK     =>  X_CLK,
    X_STROBE  =>  X_STROBE,
    X_DIN     =>  X_DIN,
    X_DOUT    =>  X_DOUT,
    X_DAV     =>  X_DAV,
    X_BSY     =>  X_BSY,

    DSR_clk   => DSR_clk,
    OSR_clk   => OSR_clk,
    DSR_clk_out => DSR_clk_out,
    OSR_clk_out => OSR_clk_out

  );


  RunVarious: process
  begin
    MLDL_SR   <= '1';     -- default, SR from SRAM
    SR_SET(0) <= '1';
    SR_SET(1) <= '1';
    MLDL_DIS  <= '1';     -- MLDL enable
    wait;
  end process;


  RunDSR_clk: process
  -- connect DSR_clk_out with DSR_clk
  begin
    DSR_clk <= DSR_clk_out;
    loop
      wait until DSR_clk_out = '0';
      DSR_clk <= '0';
      wait until DSR_clk_out = '1';
      DSR_clk <= '1';
    end loop;
  end process;

  RunOSR_clk: process
  -- connect DSR_clk_out with DSR_clk
  begin
    OSR_clk <= OSR_clk_out;
    loop
      wait until OSR_clk_out = '0';
      OSR_clk <= '0';
      wait until OSR_clk_out = '1';
    OSR_clk <= '1';
    end loop;
  end process;


  ReadRunStim : process
-- Renewed RunStim process to allow for IO stimuli as well
  variable Hstim: hp41stim;
  variable Xstim: IOstim;
  variable add: natural;
  variable cycle: natural;
  variable L: line;
  variable i, j: natural;
  variable c, c1, c2, c3, c4, c5: character;
  variable chrs : data_char;
  variable temp : nybble;
  file stim_file : TEXT open read_mode is "hp41_stim.rom";


  procedure AnalyzeHPline is
  begin
   read (L, c);               -- skip space
   read (L, c1);              -- read ISA address
    read (L, c2);
    read (L, c3);
    read (L, c4);
    case c is
      when 'X' =>         -- assume c2, c3, c4 are all X
        Hstim.stim_Addr := "XXXXXXXXXXXXXXXX";
      when 'Z' =>         -- assume c2, c3, c4 are all Z
        Hstim.stim_Addr := "ZZZZZZZZZZZZZZZZ";
      when 'U' =>         -- assume c2, c3, c4 are all U
        Hstim.stim_Addr := "UUUUUUUUUUUUUUUU";
      when others =>
        Hstim.stim_Addr := conv_std_logic_vector( 4096 * hex2bin(c1) +
                                     256 * hex2bin(c2) +
                                      16 * hex2bin(c3) +
                                           hex2bin(c4), 16);
    end case;
    read (L, c);          -- read delimiting space (or tab)
    read (L, c1);         -- read ISA instruction
    read (L, c2);
    read (L, c3);
    case c1 is
      when 'X' =>         -- assume c2, c3 are all X
        Hstim.stim_Instr := "XXXXXXXXXX";
      when 'Z' =>         -- assume c2, c3 are all Z
        Hstim.stim_Instr := "ZZZZZZZZZZ";
      when 'U' =>         -- assume c2, c3 are all Z
        Hstim.stim_Instr := "UUUUUUUUUU";
      when others =>
        Hstim.stim_Instr := conv_std_logic_vector(  256 * hex2bin(c1) +
                                       16 * hex2bin(c2) +
                                            hex2bin(c3), 10);
    end case;
    read (L, c);          -- read delimiting space (or tab)
    read (L, c1);         -- read SYNC instruction, only 0 or 1
    case c1 is
      when '1' =>         -- assume c2, c3 are all X
        Hstim.stim_SYNC := '1';   -- drive SYNC
      when others =>
        Hstim.stim_SYNC := '0';   -- do not drive SYNC (keep low)
    end case;
    read (L, c);          -- read delimiting space (or tab)
    read (L, chrs(13));     -- first character, sign
    read (L, c);          -- delimiting dot
    for i in 12 downto 3 loop
      read (L, chrs(i));    -- read mantissa
    end loop;
    read (L, c);          -- delimiting dot
    read (L, chrs(2));      -- exponent sign
    read (L, c);          -- delimiting dot
    read (L, chrs(1));      -- exponent MSD
    read (L, chrs(0));      -- exponent LSD
    case chrs(13) is
      when 'X' =>         -- assume c2, c3 are all X
        for i in 0 to 55 loop
          Hstim.stim_DATA(i) := 'X';
        end loop;
      when 'Z' =>         -- assume c2, c3 are all Z
        for i in 0 to 55 loop
          Hstim.stim_DATA(i) := 'Z';
        end loop;
      when 'U' =>         -- assume c2, c3 are all U
        for i in 0 to 55 loop
          Hstim.stim_DATA(i) := 'U';
        end loop;
      when others =>          -- decode hex character
        for i in 0 to 13 loop
        temp := conv_std_logic_vector(hex2bin(chrs(i)),4);
          for j in 0 to 3 loop
            Hstim.stim_DATA(i*4+j) := temp(j);
          end loop;
        end loop;
    end case;
  end;

  procedure AnalyzeIOline is
  begin
    read(L, c);          -- skip space
    read (L, c1);        -- read Mode (always 3 characters)
    read (L, c2);
    read (L, c3);
    case c1 is
       when 'S' =>                   -- S_L or S_H
          case c3 is
             when  'L' =>                     -- S_L
                Xstim.stim_Mode := 0;
             when 'H' =>                      -- S_H
                Xstim.stim_Mode := 1;
             when others =>                   -- set error, must prevent this!
                Xstim.stim_Mode := 99;
          end case;
       when 'C' =>                   -- C_L or C_H
          case c3 is
             when  'L' =>                     -- S_L
                Xstim.stim_Mode := 2;
             when 'H' =>                      -- S_H
                Xstim.stim_Mode := 3;
             when others =>                   -- set error, must prevent this!
                Xstim.stim_Mode := 99;
          end case;
       when 'W' =>                   -- WAI
          Xstim.stim_Mode := 4;
      when 'D' =>                    -- DAW or DAR
          case c3 is
             when 'W' =>                      -- DAW
               Xstim.stim_Mode := 5;
               -- read rest of line
               read (L, c);            -- skip space
               read (L, c1);           -- this is Data
               read (L, c2);
               read (L, c3);
               Xstim.stim_Data := conv_std_logic_vector(256 * hex2bin(c1) +
                                                        16 * hex2bin(c2) +
                                                             hex2bin(c3), 10);
               read (L, c);            -- skip dot
               read (L, c1);           -- read Command
               Xstim.stim_Cmd := conv_std_logic_vector(hex2bin(c1),4);
               read (L, c);            -- skip dot
               read (L, c1);           -- read Address
               read (L, c2);
               read (L, c3);
               read (L, c4);
               read (L, c5);
               Xstim.stim_Addr := conv_std_logic_vector(65536 * hex2bin(c1) +
                                                         4096 * hex2bin(c2) +
                                                          256 * hex2bin(c3) +
                                                           16 * hex2bin(c4) +
                                                                hex2bin(c5), 20);
             when 'R' =>                      -- DAR
               Xstim.stim_Mode := 6;
                -- read rest of line
               read (L, c);            -- skip space
               read (L, c1);           -- this is Command
               Xstim.stim_Cmd := conv_std_logic_vector(hex2bin(c1),4);
               read (L, c);            -- skip dot
               read (L, c1);           -- read Address
               read (L, c2);
               read (L, c3);
               read (L, c4);
               read (L, c5);
               Xstim.stim_Addr := conv_std_logic_vector(65536 * hex2bin(c1) +
                                                         4096 * hex2bin(c2) +
                                                          256 * hex2bin(c3) +
                                                           16 * hex2bin(c4) +
                                                                hex2bin(c5), 20);
             when others =>                   -- set error, must prevent this!
               Xstim.stim_Mode := 99;
          end case;
       when 'I' =>                         -- IOR or IOW
          case c2 is
             when 'W' =>                      -- IOW
                Xstim.stim_Mode := 7;
                -- read rest of line
              read (L, c);            -- skip space
                read (L, c1);           -- this is Data
                read (L, c2);
                Xstim.stim_Data := conv_std_logic_vector(16 * hex2bin(c1) +
                                                             hex2bin(c2), 8);
             when 'R' =>                      -- IOR
                Xstim.stim_Mode := 8;
             when others =>                   -- set error, must prevent this!
                Xstim.stim_Mode := 99;
          end case;

       when 'E' =>                  -- ENL or ENH
        case c2 is
          when 'L' =>                 -- ENL
            Xstim.stim_Mode := 9;
          when 'H' =>                 -- ENH
            Xstim.stim_Mode := 10;
          when others =>                      -- set error, must prevent this!
              Xstim.stim_Mode := 99;

          end case;

       when others =>                      -- set error, must prevent this!
          Xstim.stim_Mode := 99;
    end case;
  end;


  procedure ExecuteHPline is
  -- Improved version for clarity
  -- wait until PHASE 46, start with SYNC and Instruction\
  -- Note that we are using the wrong DATA at D43..D55
  -- this DATA is not used in the MLDL, so ignore this

  begin

    wait until CLK02'event and CLK02 = '1' and phase_counter = 46;


    DATA <= Hstim.stim_DATA(44);
    ISA <= Hstim.stim_Instr(0);
    SYNC <= Hstim.stim_SYNC;
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(45);
    ISA <= Hstim.stim_Instr(1);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(46);
    ISA <= Hstim.stim_Instr(2);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(47);
    ISA <= Hstim.stim_Instr(3);
    SYNC <= '0'; wait for CLK_HIGH;   -- SYNC dips at PHASE 49
    SYNC <= Hstim.stim_SYNC;
    wait for CLK_PERIOD - CLK_HIGH;

    DATA <= Hstim.stim_DATA(48);
    ISA <= Hstim.stim_Instr(4);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(49);
    ISA <= Hstim.stim_Instr(5);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(50);
    ISA <= Hstim.stim_Instr(6);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(51);
    ISA <= Hstim.stim_Instr(7);
    SYNC <= '0'; wait for CLK_HIGH;   -- SYNC dips at PHASE 53
    SYNC <= Hstim.stim_SYNC;
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(52);
    ISA <= Hstim.stim_Instr(8);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(53);
    ISA <= Hstim.stim_Instr(9);
    wait for CLK_PERIOD;

    -- end of Instruction part

    SYNC <= '0';
    ISA <= 'Z';

    DATA <= Hstim.stim_DATA(54);
    wait for CLK_PERIOD;

    DATA <= Hstim.stim_DATA(55);
    wait for CLK_PERIOD;

    for i in 2 to 15 loop
      DATA <= Hstim.stim_DATA(i-2);
      wait for CLK_PERIOD;
    end loop;

    -- at PHASE 16 we also need to start driving ISA ADDRESS
    for i in 16 to 31 loop
      ISA <= Hstim.stim_Addr(i-16);
      DATA <= Hstim.stim_DATA(i-2);
      wait for CLK_PERIOD;
    end loop;

    -- at phase 32 we can stop driving ISA ADDRESS
    ISA <= 'Z';

    DATA <= Hstim.stim_DATA(30);

    for i in 33 to 45 loop
      wait for CLK_PERIOD;
      DATA <= Hstim.stim_DATA(i-2);
--      wait for CLK_PERIOD;
    end loop;

  end;


  procedure WriteData is
  begin
     -- send Data
    for i in 0 to 9 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Data(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
    -- send Command
    for i in 0 to 1 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Cmd(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
    -- send Address
    for i in 0 to 19 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Addr(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
    wait for 50 ns;
    -- assert X_STROBE
    X_STROBE <= '0';
    wait for 50 ns;
    -- assert X_CLK
    X_CLK <= '0';
    wait for 100 ns;
    X_CLK <= '1';
    wait for 50 ns;
    X_STROBE <= '1';
  end;

  procedure ReadData is
  begin
     -- send Command
    for i in 0 to 1 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Cmd(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
    -- send Address
    for i in 0 to 19 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Addr(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
    wait for 50 ns;
    -- assert X_STROBE
    X_STROBE <= '0';
    wait for 50 ns;
    -- assert X_CLK
    X_CLK <= '0';
    wait for 100 ns;
    X_CLK <= '1';
    wait for 50 ns;
    X_STROBE <= '1';
    wait for 50 ns;
    X_CLK <= '0';
      for i in 0 to 9 loop
        X_CLK <= '1';
        wait for 50 ns;
        X_CLK <= '0';
        wait for 50 ns;
      end loop;
      X_CLK <= '1';
  end;

  procedure WriteIO is
  begin
    for i in 0 to 7 loop
      X_CLK <= '0';
      X_DIN <= Xstim.stim_Data(i);
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
  end;

  procedure ReadIO is
begin
    for i in 0 to 7 loop
      X_CLK <= '0';
      wait for 50 ns;
      X_CLK <= '1';
      wait for 50 ns;
    end loop;
  end;

  procedure ExecuteIOline is
   variable i : natural;
  begin
     case xstim.stim_Mode is
        when 0 =>                        -- S_L
           X_STROBE <= '0';
        when 1 =>                        -- S_H
           X_STROBE <= '1';
        when 2 =>                        -- C_L
           X_CLK <= '0';
        when 3 =>                        -- C_H
           X_CLK <= '1';
        when 4 =>                        -- WAT
           wait for X_delay;
        when 5 =>                        -- DAW
           WriteData;
        when 6 =>                        -- DAR
           ReadData;
        when 7 =>                        -- IOW
            WriteIO;
        when 8 =>                        -- IOR
            ReadIO;
        when 9 =>                        -- ENL
            X_EN <= '0';
         when 10 =>               -- ENH
          X_EN <= '1';
        when others =>             -- should bot happen, ignore
            NULL;
     end case;
  end;


  begin
     -- Initialize relevant signals in default mode
    ISA <= 'Z';
    DATA <= 'Z';
    SYNC <= '1';
    X_EN <= '1';
    X_STROBE <= '1';
    X_CLK <= '1';           -- driven in Run_OCLK
      X_DIN <= '1';
--      H_BSY <= '0';
--      H_DAV <= '0';
     -- Start by reading one line of the stimfile
     while not endfile(stim_file) loop
        readline(stim_file, L);
        read(L, c1);

        case c1 is
           when 'H' =>    -- HP41 Bus Cycle
              AnalyzeHPline;
              ExecuteHPline;
           when 'X' =>    -- IO Cycle
              AnalyzeIOline;
              ExecuteIOLine;
            when others =>
        end case;
     end loop;

     assert false report "STIMULUS FILE EXHAUSTED";
     wait;
  end process;   -- ReadRunStim



  RunPWO : process
-- Control PWO signal
  begin
    PWO <= '0';
    wait for 2*CLK_PERIOD;
    PWO <= '1';
    wait;
  end process;


  RunClock : process
-- drives the HP41 clocks, but accelerated
  begin
    if PWO = '1' then
      CLK02 <= '1';
    end if;
    wait for CLK_HIGH;
    CLK02 <= '0';
    wait for CLK_LAG-CLK_HIGH;
    if PWO = '1' then
      CLK01 <= '1';
    end if;
    wait for CLK_HIGH;
    CLK01 <= '0';
    wait for CLK_LAG-CLK_HIGH;
  end process;


  ClockCount: process (phase_counter, PWO, CLK01)
-- keeps track of the clock phase
-- resets on PWO active low (PWO needs to be implemented)
  begin
    if PWO = '0' then
      phase_counter <= 0;
      cycle_counter <= 0;
    elsif CLK01'event and CLK01 = '1' then
      if phase_counter = 55 then
        phase_counter <= 0;
        cycle_counter <= cycle_counter + 1;
      else
        phase_counter <= phase_counter + 1;
      end if;
    end if;
  end process;


  EPROM: process
-- simulates the FLASH EPROM

  variable Flash_eprom  : Flash_memory;
  variable add: natural;
  variable L: line;
  variable i: natural;
  variable c, c1, c2, c3, c4: character;
  file Flash_file : TEXT open read_mode is "Flash.rom";

    procedure load_rom is
    begin
      -- initialize memory space with $FF
      for add in addr_0k to addr_64k loop
        Flash_eprom(add) := "1111111111111111";
      end loop;
      -- read Flash data file
      while not endfile(Flash_file) loop
        readline(Flash_file, L);
        read (L, c1);
        while c1 = '-' loop     -- skip comments
          readline(Flash_file, L);
          read (L, c1);
        end loop;
        read (L, c2);         -- read address
        read (L, c3);
        read (L, c4);
        add := 4096 * hex2bin(c1) + 256 * hex2bin(c2) + 16 * hex2bin(c3) + hex2bin(c4);
        read (L, c);          -- read delimiting space
        read (L, c1);         -- read data
        read (L, c2);
        read (l, c3);
        Flash_eprom(add) := conv_std_logic_vector(256 * hex2bin(c1) +
                                                   16 * hex2bin(c2) +
                                                        hex2bin(c3), 16);
      end loop;
    end load_rom;


  begin  -- process eprom
  load_rom;     -- read FLASH data from file
    XD <= "ZZZZZZZZZZZZZZZZ";
    loop
      wait until FLASH_CE'event and FLASH_CE = '0';
      wait for FLASH_access;    -- assume slow EPROMs (NOTE factor 10 in timing!!)
      -- decode address on bus
      -- only 16 address bits are decoded as we use only 64kbyte devices
      if FLASH_CE = '0' then
        if FLASH_OE = '0' then    -- READ access
          add := conv_integer(XA and "00001111111111111111");
          XD <= Flash_eprom(add);
        elsif FLASH_WE = '0' then
          -- this is a WRITE access
          -- we do not really have to do anything ...
        end if;
      end if;
      -- wait for EPROM to be deselected
      wait until FLASH_CE = '1';
      XD <= "ZZZZZZZZZZZZZZZZ";
    end loop;
  end process;  -- eprom


  SRAM: process
-- simulates the FLASH EPROM and SRAM

  variable SRAM_eprom : SRAM_memory;
  variable add: natural;
  variable L: line;
  variable i: natural;
  variable c, c1, c2, c3, c4: character;
  file SRAM_file : TEXT open read_mode is "sram.rom";

    procedure load_sram is
    begin
      -- initialize memory space with $FF
      for add in addr_0k to addr_64k loop
        SRAM_eprom(add) := "1111111111111111";
      end loop;
      -- read SRAM data file
      while not endfile(SRAM_file) loop
        readline(SRAM_file, L);
        read (L, c1);
        while c1 = '-' loop     -- skip comments
          readline(SRAM_file, L);
          read (L, c1);
        end loop;
        read (L, c2);         -- read address
        read (L, c3);
        read (L, c4);
        add := 4096 * hex2bin(c1) + 256 * hex2bin(c2) + 16 * hex2bin(c3) + hex2bin(c4);
        read (L, c);          -- read delimiting space
        read (L, c1);         -- read data
        read (L, c2);
        read (L, c3);
        SRAM_eprom(add) := conv_std_logic_vector(256 * hex2bin(c1) +
                                                  16 * hex2bin(c2) +
                                                       hex2bin(c3), 16);
      end loop;
    end load_sram;

  begin  -- process sram
    load_sram;      -- read SRAM data from file
    XD <= "ZZZZZZZZZZZZZZZZ";
    loop
      wait until SRAM_CE'event and SRAM_CE = '0';
      wait for SRAM_access;   -- assume slow EPROMs (NOTE factor 10 in timing!!)
      -- decode address on bus
      -- only 16 address bits are decoded as we use only 64kbyte devices
      if SRAM_CE = '0' then
        if SRAM_OE = '0' then
          add := conv_integer(XA and "00001111111111111111");
          XD <= SRAM_eprom(add);
        elsif SRAM_WE = '0' then
          -- writes are not really done here ....
          -- SRAM_eprom(add) <= conv_std_logic_vector(XD);
          --  NEED TO CHECK TYPE CAST HERE!!!!
          -- put option here to write to SRAM file on a stimulus ??
        end if;

      end if;
      -- wait for SRAM to be deselected
      wait until SRAM_CE = '1';
      XD <= "ZZZZZZZZZZZZZZZZ";
    end loop;
  end process;  -- sram

END;