-------------------------------------------------------------------------------
--    Copyright (c) 2005  Meindert Kuipers, Netherlands                      --
--    meindert@kuipers.to                www.kuipers.to                      --
--                                                                           --
-- This program is free software; you can redistribute it and/or             --
-- modify it under the terms of the GNU General Public License               --
-- as published by the Free Software Foundation; either version 2            --
-- of the License, or (at your option) any later version.                    --
--                                                                           --
-- This program is distributed in the hope that it will be useful,           --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of            --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             --
-- GNU General Public License for more details.                              --
--                                                                           --
-- You should have received a copy of the GNU General Public License         --
-- along with this program; if not, write to the Free Software               --
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA--
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-------------------------------------------------------------------------------
--             SmartROM.vhdl                                                 --
--  Ver  Date     Description                                                --
--  1.00 Dec 2004 First release version                                      --
-------------------------------------------------------------------------------

entity SmartROM is Port (

-- HP41 Bus Signals
  SYNC,         -- HP41 Signal: Indicates start of bus cycle
  CLK01,        -- HP41 Signal: Rising edge clocks data
  CLK02,        -- HP41 Signal: Count clock cycle state and output clock
  DATA,         -- HP41 Signal: Data from C-register
  ISA,          -- HP41 Signal: Address and Instruction line
  PWO:          -- HP41 Signal: Power On line
                in std_logic;

-- HP41 Output Control Signals
  ISA_out,      -- ISA Output to driver
  ISA_OE,       -- ISA Output Enable for driver
  DATA_out,     -- DATA Output to driver
  DATA_OE,      -- DATA Output Enable for driver
  SYNC_out,     -- SYNC Output to driver
  SYNC_OE,      -- SYNC Output Enable fordriver
  FI_Out,       -- Peripheral Flag Output
  FI_OE:        -- Peripheral Flag Output Enable
                out std_logic;

-- Memory Control Signals
  XA:           -- 19-bits External Address Bus for FLASH and SRAM
                out std_logic_vector(19 downto 00);
  XD:           -- 10-bit external Data Bus
                inout std_logic_vector(15 downto 00);
--  XXD:          -- spare Data bits, defined as input with weak pullup
--                inout std_logic_vector(15 downto 10);

  FLASH_CE,     -- FLASH Chip Enable (active low)
  FLASH_OE,     -- FLASH Output Enable (active low)
  FLASH_WE,     -- FLASH Write Enable (active low signal)
  SRAM_CE1,     -- SRAM Chip Enable 1 (active low signal)
  SRAM_CE2,     -- SRAM Chip Enable 2 (active high signal)
  SRAM_OE,      -- SRAM Output Enable (active low signal)
  SRAM_WE,      -- SRAM Write Enable (active low signal)
  MEM_RESET:    -- FLASH reset signal
                out std_logic;
  RY_BY:        -- FLASH RY/BY# signal
                in std_logic;

-- Utility Signals connected to DIPSWITCH
  MLDL_SR:      in std_logic;  -- Status Registers from SRAM (1) or FLASH (0)
  SR_SET:       in std_logic_vector(1 downto 0);
                               -- select 1 of 4 Settings Register Sets
  MLDL_DIS:     in std_logic;  -- disable MLDL output to HP41

-- External I/O Interface Signals
  X_EN,                         -- Enable eXternal Interface
  X_CLK,                        -- eXternal Interface Clock   on JTAG-TCK!
  X_STROBE,                     -- eXternal Strobe            on JTAG-TMS!
  X_DIN:        in std_logic;   -- eXternal Input             on JTAG-TDI!
  X_DOUT:       out std_logic;  -- eXternal Output            on JTAG-TDO!
  X_DAV,                        -- Data Available in MLDL
  X_BSY:        out std_logic;  -- MLDL Busy on current data

  DSR_clk,
  OSR_clk:      in std_logic;  -- to use global clocks here
  DSR_clk_out,
  OSR_clk_out:  out std_logic

  );

  end SmartROM;

architecture Behavioral of SmartROM is

signal  ISR:      std_logic_vector(15 downto 00);   -- ISA Shift Register
alias   I:        std_logic_vector(09 downto 00) is ISR(09 downto 00);
alias   EBR_A:    std_logic_vector(02 downto 00) is ISR(15 downto 13);
alias   EBR_D:    std_logic_vector(01 downto 00) is ISR(07 downto 06);

signal  OSR:      std_logic_vector(09 downto 00);   -- ISA Output Shift Register
signal  ISA_O:    std_logic;                        -- ISA Output internal

signal  DSR:      std_logic_vector(31 downto 00);   -- DATA Shift Register
signal  XSR:      std_logic_vector(05 downto 00);   -- DATA Shift Register

-- HP41 I/O Register
signal  IOR:      std_logic_vector(15 downto 00);   -- HP41 I/O Register
signal  H_DAV:    std_logic;    -- I/O status bit: Data Available for reading
signal  H_BSY:    std_logic;    -- I/O status bit: Data Pending in I/O
signal  HDAV_clk: std_logic;    -- clock to control H_DAV register
signal  HBSY_clk: std_logic;    -- clock to control H_BSY register

-- temp storage Settings Register
signal  SR:       std_logic_vector(9 downto 0);    -- ROM Address Translation

-- Phase Counter and clocks
signal  ph_count:   std_logic_vector(5 downto 0);
signal  ph_reset,
        t1_reset,
        t2_reset:   std_logic;
signal  IOR_clk:    std_logic;      -- Clock for IOR

signal  DATA_Start,                 -- Indicates start of DATA
        DATA_shift: std_logic;      -- High when DATA is present

signal  SRAM_CE:    std_logic;     -- Internal SRAM CE

-- Main MLDL State Machine
type MLDLState is (
      Idle,                   -- nothing to do ....
      MLDL_disable,           -- MLDL is disabled and ready for I/O
      ROMReadSR,              -- Read Status Register for FLASH/SRAM Access
      ROMRead,                -- Read data for FLASH/SRAM access
      ShiftInISA,             -- Regular ISA ISR read, no OSR shifting
      ShiftOutISA,            -- Move data in OSR out to ISA (and read back!)
      DecodeInstr,            -- Decode Instruction in ISR
      ReadAddrISA,            -- Read Address on ISA
      WROMReadSR,             -- Read SR for WROM preparation
      WROMWrite,              -- Write for WROM
      IORead,                 -- Read from IO Register
      IOWrite                 -- Write to IO Register
    );
signal MLDLStat, next_MLDLStat : MLDLState;

-- Instruction Decoder state machine
type  I_State is (
      I_Idle,           -- no instruction recognized
      I_WROM,           -- instruction: WROM      040
      I_FETCH,          --              FETCH S&X 030
      I_ENBANK          --              ENBANKx   100/180/140/1C0
      );
signal IStat: I_State;

-- registers for Bank Switching
type   Bank_type is array(7 downto 0) of std_logic_vector (1 downto 0);
signal EBR: Bank_type := ("00","00","00","00","00","00","00","00");
signal B: std_logic_vector (1 downto 0);  -- temp storage for Active Bank


begin

-------------------------------------------------------------------------------
--             PHASE COUNTER                                                 --
-------------------------------------------------------------------------------

sync_detect: process(SYNC, CLK01, CLK02, t1_reset, t2_reset)
-- Generates the actual counter reset after detecting SYNC sequence
-- and on phase 67 (octal) if no SYNC is present
begin
  if (CLK01'event and CLK01 = '1') then
    t1_reset <= SYNC;
    if ((ph_count = "110111") or t1_reset = '1') then     -- Phase 55
      t2_reset <= '1';
    else
      t2_reset <= '0';
    end if;
  end if;
  ph_reset <= (t2_reset and CLK02 and not t1_reset);
end process;

phase_counter: process(CLK02, PWO, ph_reset)
-- 6-bit phase counter
begin
  if ph_reset = '1' or (PWO = '0') then
    ph_count <= "000000";                  -- reset
  elsif (CLK02'event and CLK02 = '1') then
    ph_count <= ph_count + 1;
  end if;
end process;

-------------------------------------------------------------------------------
--             TIMING SIGNAL GENERATOR                                       --
-------------------------------------------------------------------------------

ISA_timing: process(ph_count)
begin
  if (ph_count = "000001") then             -- phase 1
    DATA_start <= '1';
  else
    DATA_Start <= '0';
  end if;
end process;

DATA_FF: process (DATA_start, IStat, ph_count)
-- DATA_shift indicates when interesting DATA is present on DATA
begin
  if ((ph_count = "100001") or IStat = I_Idle or IStat = I_ENBANK) then
     -- phase 33 is start of interesting DATA
    DATA_shift <= '0';
  elsif (DATA_start'event and DATA_start = '1') then
    DATA_shift <= '1';
  end if;
end process;

IOR_clock: process(X_STROBE, X_CLK, CLK01, MLDLStat, SR, ISR)
-- generates the clock for the IO Register
begin
  if ((MLDLStat = WROMWrite) and (SR(07) = '1') and
       ISR(11) = '1' and ISR(10) = '0' ) then
     -- parallel loading of IOR when writing to I/O Register
     --   this is during a WROM, SR(07) indicates IO Bank
     --   ISR(11) =0 and ISR(10) = 0: range is $x800-$xBFF
    IOR_clk <= CLK01;
  else
    IOR_clk <= X_CLK;
  end if;
end process;


-------------------------------------------------------------------------------
--           MLDL2000 MAIN STATE MACHINE                                     --
-------------------------------------------------------------------------------

MLDL_state_machine: process (MLDLStat, X_EN, IStat, DATA_Shift, SYNC, SR, ph_count)
begin
  next_MLDLStat <= Idle;                -- default state value
  case MLDLStat is

    when Idle =>                        -- Idle state, wait with nothing to do
      if (ph_count = "001110") then     -- phase 14
        next_MLDLStat <= ReadAddrISA;
      end if;
      if (ph_count = "101100") then     -- ISA Instruction time, phase 44
        if ((MLDL_DIS = '0') and (SR(09) = '0')) then   -- valid ROM access, OSR shift out
          next_MLDLStat <= ShiftOutISA;
        else
          next_MLDLStat <= ShiftInISA;
        end if;
      end if;
      if X_EN = '0' then
        next_MLDLStat <= MLDL_Disable;
      end if;
      if ((ph_count = "100001") and IStat = I_WROM) then   -- phase 33
        -- Add code here to prevent SR access when in pages 0, 1, 2 to save power
        next_MLDLStat <= WROMReadSR;
      end if;
      if  ((ph_count = "100001") and IStat = I_FETCH) then  -- phase 33
        -- Add code here to prevent SR access when in pages 0, 1, 2 to save power
        next_MLDLStat <= ROMReadSR;
      end if;

    when MLDL_disable =>              -- MLDL is disabled and ready for I/O
      if X_EN = '1' then
        next_MLDLStat <= Idle;
      else
        next_MLDLStat <= MLDL_disable;
      end if;

    when ROMReadSR =>                 -- Read Status Register for FLASH/SRAM Access
      if SR(09) = '1' then            -- if disabled bank ....
        next_MLDLStat   <= Idle;
      elsif SR(08) = '1' and SR(07) = '1' then     -- This is an I/O access
         next_MLDLStat <= IORead;
      else
        next_MLDLStat <= ROMRead;
      end if;

    when ROMRead =>                   -- Read L8 data for FLASH/SRAM access
      next_MLDLStat <= Idle;

    when ShiftInISA =>                -- Regular ISA ISR read, no OSR shifting
      if (ph_count = "110110") then   -- phase 27
        if SYNC = '1' then
          next_MLDLStat <= DecodeInstr;
        else
          next_MLDLStat <= Idle;
        end if;
      else
        next_MLDLStat <= ShiftInISA;
      end if;

    when ShiftOutISA =>               -- Move data in OSR out to ISA and read back
      if (ph_count = "110110") then   -- phase 27
        if SYNC = '1' then
          next_MLDLStat <= DecodeInstr;
        else
          next_MLDLStat <= Idle;
        end if;
      else
        next_MLDLStat <= ShiftOutISA;
      end if;

    when DecodeInstr =>               -- Decode Instruction in ISR
      next_MLDLStat <= Idle;

    when ReadAddrISA =>               -- Read Address on ISA
      if (ph_count = "011110") then   -- phase 30
        if DATA_Shift = '1' then
          next_MLDLStat <= Idle;
        else
          next_MLDLStat <= ROMReadSR;
        end if;
      else
        next_MLDLStat <= ReadAddrISA;
      end if;

    when WROMReadSR =>                -- Read SR for WROM preparation
      if (SR(08) = '0' or SR(06) = '1') then  -- if Write protected or FLASH
        next_MLDLStat <= ROMReadSR;
      else
       if SR(07) = '0' then
          next_MLDLStat <= WROMWrite;
        else
          next_MLDLStat <= IOWrite;
        end if;
      end if;

    when WROMWrite  =>                --  Write for WROM
      next_MLDLStat <= ROMReadSR;

    when IORead =>
      next_MLDLStat <= Idle;

    when IOWrite =>
      next_MLDLStat <= Idle;

  end case;
end process;

process (CLK02, ph_reset, X_EN, next_MLDLStat)
begin
  if X_EN = '0' then
    MLDLStat <= MLDL_disable;
  elsif (CLK02'event and CLK02 = '1') then
    MLDLStat <= next_MLDLStat;
  end if ;
end process ;

-------------------------------------------------------------------------------
--           INSTRUCTION DETECTOR                                            --
-------------------------------------------------------------------------------

I_detect: process(I, MLDLStat, ph_count)
begin
-- only when a SYNC has been present
  if (MLDLStat = DecodeInstr) then
   if I = "0001000000" then             --  WROM    $040
      IStat <= I_WROM;
    elsif I = "1100110000" then         --  FETCH   $330
       IStat <= I_FETCH;
    elsif I(5 downto 0) = "000000" and I(9) = '0' and I(8) = '1' then
      IStat <= I_ENBANK;                --  ENBANKx:  $100/$180/$140/$1C0
    else
      IStat <= I_Idle;
    end if;
  elsif (ph_count = "101100") then      -- phase 44
    IStat <= I_Idle;
  end if;
end process;

--    Instructions for future upgrades:
--       $270     RAMSLCT        Select RAM Block in C[2..0]
--       $2F0     WRITDATA       Writes C[13..0] to selected RAM register
--       $038     READDATA       reads selected RAM register to C[13..0]
--       $028     WRIT 0..F      write to RAM register in slected block
--       $038     READ 0..F      read from RAM register in selected block
--       $3F0     PRPHSLCT       Select Peripheral in C[1..0]
--       $024     SELPF 0..F     alow peripheral to take control
--       $1F0     WPTOG          Toggle Write Protection (HEPAX)
--       $030     BLKMOV         Move page (HEPAX)


-------------------------------------------------------------------------------
--           ENABLED BANK REGISTERS                                          --
-------------------------------------------------------------------------------

EBR_control: process(CLK01, IStat, MLDLStat, ISR, EBR_D)
-- Controls access to Enabled Bank Registers
begin
  if (CLK01'event and CLK01 = '1') then
    if ((IStat = I_ENBANK) and (MLDLStat = DecodeInstr)) then
      -- works in all Pages, Even and Odd pages share the same EBR!
      EBR(conv_integer(EBR_A)) <= EBR_D;
    end if;
  end if;
end process;


EBR_output: process(EBR, EBR_A)
-- Controls the EBR output for the correct page selection in SR access
begin
  B <= EBR(conv_integer(EBR_A));
end process;

-------------------------------------------------------------------------------
--           ISA INPUT SHIFT REGISTER                                        --
-------------------------------------------------------------------------------

ISA_input: process(CLK01, ISA, ISR, MLDLStat)
-- ISA INPUT SHIFT REGISTER
begin
  if (CLK01'event and CLK01 = '1') then
    -- this construction is chosen to keep the page address in ISR
    -- to facilitate ENBANK instructions
    if (MLDLStat = ReadAddrISA) then
      ISR(00) <= ISR(01); ISR(01) <= ISR(02);
      ISR(02) <= ISR(03); ISR(03) <= ISR(04);
      ISR(04) <= ISR(05); ISR(05) <= ISR(06);
      ISR(06) <= ISR(07); ISR(07) <= ISR(08);
      ISR(08) <= ISR(09); ISR(09) <= ISR(10);
      ISR(10) <= ISR(11); ISR(11) <= ISR(12);
      ISR(12) <= ISR(13); ISR(13) <= ISR(14);
      ISR(14) <= ISR(15); ISR(15) <= ISA;
    elsif ((MLDLStat = ShiftInISA) or (MLDLStat = ShiftOutISA)) then
      ISR(00) <= ISR(01); ISR(01) <= ISR(02);
      ISR(02) <= ISR(03); ISR(03) <= ISR(04);
      ISR(04) <= ISR(05); ISR(05) <= ISR(06);
      ISR(06) <= ISR(07); ISR(07) <= ISR(08);
      ISR(08) <= ISR(09); ISR(09) <= ISA;
    end if;
  end if;
end process;


-------------------------------------------------------------------------------
--           DATA INPUT SHIFT REGISTER                                       --
-------------------------------------------------------------------------------

DATA_clock: process(CLK01, DATA_shift, X_EN, X_STROBE, X_CLK)
-- controls clock for Data shifter
begin
   -- DSR_clk_out is an output, connect externally to DSR_clk !!!
   -- DSR_clk should be a global clock for easier routing
  if (X_EN = '1' and DATA_shift = '1') then   -- input of DATA
    DSR_clk_out <= CLK01;
  elsif (X_EN = '0' and X_STROBE = '1') then  -- no clocks when X_STROBE = '0'
    DSR_clk_out <= X_CLK;
  else                                        -- no clocks at all
    DSR_clk_out <= '0';
  end if;
end process;


DATA_input: process(DATA, DSR, DSR_clk, X_EN, X_DIN)
-- DATA INPUT SHIFT REGISTER
begin
  if (DSR_clk'event and DSR_clk = '1') then   -- regular Data Shift Register
    XSR(00) <= XSR(01);   XSR(01) <= XSR(02);   -- for 16-bit memory access
    XSR(02) <= XSR(03);   XSR(03) <= XSR(04);    
    XSR(04) <= XSR(05);   XSR(05) <= DSR(00);        
        
    DSR(00) <= DSR(01);   DSR(01) <= DSR(02);   -- normal DSR
    DSR(02) <= DSR(03);   DSR(03) <= DSR(04);
    DSR(04) <= DSR(05);   DSR(05) <= DSR(06);
    DSR(06) <= DSR(07);   DSR(07) <= DSR(08);
    DSR(08) <= DSR(09);   DSR(09) <= DSR(10);
    DSR(10) <= DSR(11);   DSR(11) <= DSR(12);
    DSR(12) <= DSR(13);   DSR(13) <= DSR(14);
    DSR(14) <= DSR(15);   DSR(15) <= DSR(16);
    DSR(16) <= DSR(17);   DSR(17) <= DSR(18);
    DSR(18) <= DSR(19);   DSR(19) <= DSR(20);
    DSR(20) <= DSR(21);   DSR(21) <= DSR(22);
    DSR(22) <= DSR(23);   DSR(23) <= DSR(24);
    DSR(24) <= DSR(25);   DSR(25) <= DSR(26);
    DSR(26) <= DSR(27);   DSR(27) <= DSR(28);
    DSR(28) <= DSR(29);   DSR(29) <= DSR(30);
    DSR(30) <= DSR(31);
    if (X_EN = '1') then
      DSR(31) <= DATA;     -- normal situation,data from HP41
    else
      DSR(31) <= X_DIN;    -- eXternal Data Input, DSR_clk <= X_CLK !!
    end if;
  end if;
end process;


-------------------------------------------------------------------------------
--           IO SHIFT REGISTER                                       --
-------------------------------------------------------------------------------
IOSR: process(IOR_clk, IOR, X_DIN, XD, X_EN, X_STROBE, DSR)
begin
  if IOR_clk'event and IOR_clk = '1' then
    if X_STROBE = '1' then
      -- serial shift register, always shift data in
      IOR(00) <= IOR(01);   IOR(01) <= IOR(02);
      IOR(02) <= IOR(03);   IOR(03) <= IOR(04);
      IOR(04) <= IOR(05);   IOR(05) <= IOR(06);
      IOR(06) <= IOR(07);
      if X_EN = '1' then       -- normal I/O, IOR_clk <= CLK01
        IOR(07) <= X_DIN;
      else     -- X_EN = '0', this is a memory read, IOR_clk <= X_CLK
        IOR(07) <= IOR(08);   IOR(08) <= IOR(09);   
        IOR(09) <= IOR(10);   IOR(10) <= IOR(11);
        IOR(11) <= IOR(12);   IOR(12) <= IOR(13);
        IOR(13) <= IOR(14);   IOR(14) <= IOR(15);
        IOR(15) <= X_DIN;
      end if;
    else  -- X_STROBE = '0'
      if X_EN = '1' then
        -- memory write operation from HP41 to IOR, IOR_clk <= CLK01
        IOR(00) <= DSR(00);   IOR(01) <= DSR(01);
        IOR(02) <= DSR(02);   IOR(03) <= DSR(03);
        IOR(04) <= DSR(04);   IOR(05) <= DSR(05);
        IOR(06) <= DSR(06);   IOR(07) <= DSR(07);
      elsif DSR(11) = '1' then
        -- X_EN = 0, and X_STROBE = 0, DSR(11) = '1', IOR_clk <= X_CLK
        -- this is a memory read from external interface so latch data from XD
        IOR(00) <= XD(00);    IOR(01) <= XD(01);
        IOR(02) <= XD(02);    IOR(03) <= XD(03);
        IOR(04) <= XD(04);    IOR(05) <= XD(05);
        IOR(06) <= XD(06);    IOR(07) <= XD(07);
        IOR(08) <= XD(08);    IOR(09) <= XD(09);
        IOR(10) <= XD(10);    IOR(11) <= XD(11);
        IOR(12) <= XD(12);    IOR(13) <= XD(13);
        IOR(14) <= XD(14);    IOR(15) <= XD(15);
      end if;
    end if;
  end if;
end process;


HDAV_clock: process(X_CLK, CLK01, MLDLStat, ISR, DSR)
-- controls the clock for the H_DAV register
-- HP41 read or writes to I/O space in the designated window
begin
  if ((MLDLStat = IOWrite and DSR(23) = '1' and DSR(22) = '0') or
      (MLDLStat = IORead  and ISR(11) = '1' and ISR(12) = '0')) then
      HDAV_clk <= CLK01;
  else
    HDAV_clk <= '0';
  end if;
end process;


HBSY_clock: process(MLDLStat, CLK01, DSR, ISR)
-- controls the clock for the H_BSY register
-- HP41 reads or writes to I/O space in the designated window
begin
  if ((MLDLStat = IOWrite and DSR(23) = '1' and DSR(22) = '0') or
      (MLDLStat = IORead  and ISR(11) = '1' and ISR(12) = '0')) then
    HBSY_clk <= CLK01;
  else
    HBSY_clk <= '0';
  end if;
end process;


Handshake_HDAV: process(X_EN, X_STROBE, X_CLK, HDAV_clk)
-- Controls H_DAV and X_BSY (NOTE that X_BSY <= H_DAV!)
begin
  if (X_EN = '1' and X_STROBE = '0' and X_CLK = '0') then
    H_DAV <= '1';
  elsif HDAV_clk'event and HDAV_clk = '0' then
    H_DAV <= '0';
  end if;
end process;


Handshake_HBSY: process(X_EN, X_STROBE, HBSY_clk, X_CLK)
-- Controls H_BSY (NOTE that X_DAV <= H_BSY!)
begin
  if (X_EN = '1' and X_STROBE = '0' and X_CLK = '0') then
     H_BSY <= '0';
  elsif HBSY_clk'event and HBSY_clk = '1' then
    H_BSY <= '1';
  end if;
end process;


Handshake: process(H_DAV, H_BSY, RY_BY, X_EN, IOR(00))
-- drives X_BSY and X_DAV lines for I/O handshaking
begin
  if X_EN = '0' then
    X_BSY <= RY_BY;    -- the FLASH EPROM READY/BUSY signal
  else
    X_BSY <= H_DAV;    -- normal use
  end if;
  X_DAV <= H_BSY;
  X_DOUT <= IOR(00);
end process;

-------------------------------------------------------------------------------
--           STATUS REGISTER                                                 --
-------------------------------------------------------------------------------

SR_Read: process(CLK01, XD, ISR, MLDLStat)
-- Read and latch the SR contents on rising edge of CLK01 and SR access
begin
--    Settings register layout:
--    SR  | 09 | 08 | 07 | 06 | 05 | 04 | 03 | 02 | 01 | 00 |
--          EN    0   IO   WP   A17  A16  A15  A14  A13  A12   for SRAM/IO
--          EN    1   A19  A18  A17  A16  A15  A14  A13  A12   for FLASH
--
--    SR(09):   1- disabled       0- enabled                     was: SR7
--    SR(08):   1- SRAM           0- FLASH                       was: SR6
--    SR(07):   1- IO Bank        0- SRAM         only if SRAM   was: SR5
--    SR(06):   1- Write Protect  0- Write Enable only if SRAM   was: SR4

  if  (CLK01'event and CLK01 = '0') then
    if (MLDLStat = ROMReadSR) or (MLDLStat = WROMReadSR) then
      SR(00) <= XD(00);       -- A12
      SR(01) <= XD(01);       -- A13
      SR(02) <= XD(02);       -- A14
      SR(03) <= XD(03);       -- A15
      SR(04) <= XD(04);       -- A16
      SR(05) <= XD(05);       -- A17
      SR(06) <= XD(06);       -- A18 if FLASH, WP/WE if SRAM
      SR(07) <= XD(07);       -- A19 if FLASH, IO/SR if SRAM
      SR(08) <= XD(08);       -- FLASH or SRAM
      SR(09) <= XD(09);       -- Enable/Disable
    end if;
  end if;
end process;


-------------------------------------------------------------------------------
--           ISA OUTPUT SHIFT REGISTER                                       --
-------------------------------------------------------------------------------

ISA_Output: process (CLK01, CLK02, MLDLStat, ISA_O, MLDL_DIS)
begin
  if (MLDLStat = ShiftOutISA) then
     -- Enable 74HCT125 Output, only if dipswitch is set
    OSR_clk_out <= CLK02;
    ISA_Out <= ISA_O;
    ISA_OE <= '0';
--    if MLDL_DIS = '1' then
--       ISA_OE <= 'Z';       -- Disabled
--    else
--       ISA_OE <= '0';       -- Enabled
--    end if;
  else
    OSR_clk_out <= CLK01;
    ISA_OE <= 'Z';        -- disable 74HCT125 output
    ISA_Out <= '0';       -- does not matter much, but watch power consumption
  end if;
end process;


ISA_OSR: process (MLDLStat, OSR, XD, OSR_clk, SR)
  -- ISA Output Shift Register
  -- read XD into the OSR and shift out ISA
subtype HA_select is std_logic_vector(2 downto 0);
begin
  if (OSR_clk'event and OSR_clk = '1') then
    case MLDLStat is

      when ShiftOutISA =>           -- need to move OSR to the ISA output
        ISA_O   <= OSR(00);     OSR(00) <= OSR(01);
        OSR(01) <= OSR(02);     OSR(02) <= OSR(03);
        OSR(03) <= OSR(04);     OSR(04) <= OSR(05);
        OSR(05) <= OSR(06);     OSR(06) <= OSR(07);
        OSR(07) <= OSR(08);     OSR(08) <= OSR(09);

      when ROMRead =>               -- regular read of ROM
        OSR(00) <= XD(00);      OSR(01) <= XD(01);
        OSR(02) <= XD(02);      OSR(03) <= XD(03);
        OSR(04) <= XD(04);      OSR(05) <= XD(05);
        OSR(06) <= XD(06);      OSR(07) <= XD(07);
        OSR(08) <= XD(08);      OSR(09) <= XD(09);

      when IORead =>
        if ISR(11) = '0' and ISR(10) = '1' then     -- Read Memory or SR
          OSR(00) <= XD(00);      OSR(01) <= XD(01);
          OSR(02) <= XD(02);      OSR(03) <= XD(03);
          OSR(04) <= XD(04);      OSR(05) <= XD(05);
          OSR(06) <= XD(06);      OSR(07) <= XD(07);
          OSR(08) <= XD(08);      OSR(09) <= XD(09);
        elsif ISR(11) = '1' and ISR(10) = '0' then  -- Read IO Register
          OSR(00) <= IOR(00);     OSR(01) <= IOR(01);
          OSR(02) <= IOR(02);     OSR(03) <= IOR(03);
          OSR(04) <= IOR(04);     OSR(05) <= IOR(05);
          OSR(06) <= IOR(06);     OSR(07) <= IOR(07);
          OSR(08) <= H_DAV;       OSR(09) <= H_BSY;
        else                                        -- outside window, return 0
          OSR <= (others => '0');
        end if;

      when others => NULL;
    end case;
  end if;

end process;


-------------------------------------------------------------------------------
--           MEMORY ACCESS LOGIC                                             --
-------------------------------------------------------------------------------

memory_access: process (MLDLStat, ISR, DSR, XSR, CLK02, SR, X_EN,
                        MLDL_SR, X_STROBE, IOR, B, SR_SET, X_CLK)
-- memory access engine, controls FLASH and SRAM
subtype HA_select is std_logic_vector(2 downto 0);

begin

  case MLDLStat is

    when ROMReadSR =>
      XA(19) <= '1';        XA(18) <= '1';
      XA(17) <= '1';        XA(16) <= '1';
      XA(15) <= '1';        XA(14) <= '1';
      XA(13) <= '1';        XA(12) <= '1';
      XA(11) <= '1';        XA(10) <= '1';
      XA(09) <= '1';        XA(08) <= '1';
      XA(07) <= not SR_SET(1);  XA(06) <= not SR_SET(0);

      XA(05) <= ISR(15);    XA(04) <= ISR(14);
      XA(03) <= ISR(13);    XA(02) <= ISR(12);

--    XA(01) <= '0';       XA(00) <= '0';           -- no Bank Switching support
      XA(01) <= B(1);      XA(00) <= B(0);          -- current active Bank

      XD <= (others => 'Z');
      SRAM_WE <= '1';     FLASH_WE <= '1';

      if MLDL_SR = '0' then                         -- SR in SRAM
        SRAM_CE <= '0';     FLASH_CE <= '1';
        SRAM_OE <= '0';     FLASH_OE <= '1';
      else                                          -- SR in FLASH
        SRAM_CE <= '1';     FLASH_CE <= '0';
        SRAM_OE <= '1';     FLASH_OE <= '0';
      end if;

    when WROMReadSR =>
      XA(19) <= '1';        XA(18) <= '1';
      XA(17) <= '1';        XA(16) <= '1';
      XA(15) <= '1';        XA(14) <= '1';
      XA(13) <= '1';        XA(12) <= '1';
      XA(11) <= '1';        XA(10) <= '1';
      XA(09) <= '1';        XA(08) <= '1';
      XA(07) <= not SR_SET(1);  XA(06) <= not SR_SET(0);

      XA(05) <= DSR(27);    XA(04) <= DSR(26);
      XA(03) <= DSR(25);    XA(02) <= DSR(24);

--    XA(01) <= '0';       XA(00) <= '0';           -- no Bank Switching support
      XA(01) <= B(1);      XA(00) <= B(0);          -- current active Bank

      XD <= (others => 'Z');
      SRAM_WE <= '1';     FLASH_WE <= '1';

      if MLDL_SR = '0' then                         -- SR in SRAM
        SRAM_CE <= '0';     FLASH_CE <= '1';
        SRAM_OE <= '0';     FLASH_OE <= '1';
      else                                          -- SR in FLASH
        SRAM_CE <= '1';     FLASH_CE <= '0';
        SRAM_OE <= '1';     FLASH_OE <= '0';
      end if;

    when ROMRead =>
      XD <= (others => 'Z');
      XA(17) <= SR(05);     XA(16) <= SR(04);
      XA(15) <= SR(03);     XA(14) <= SR(02);
      XA(13) <= SR(01);     XA(12) <= SR(00);
      XA(11) <= ISR(11);    XA(10) <= ISR(10);
      XA(09) <= ISR(09);    XA(08) <= ISR(08);
      XA(07) <= ISR(07);    XA(06) <= ISR(06);
      XA(05) <= ISR(05);    XA(04) <= ISR(04);
      XA(03) <= ISR(03);    XA(02) <= ISR(02);
      XA(01) <= ISR(01);    XA(00) <= ISR(00);

      SRAM_WE <= '1';       FLASH_WE <= '1';

      if SR(08) = '0' then                          -- FLASH is addressed
        XA(19)  <= SR(07);    XA(18)   <= SR(06);
        SRAM_CE <= '1';       FLASH_CE <= CLK02;
        SRAM_OE <= '1';       FLASH_OE <= CLK02;
      else                                          -- SRAM is addressed
        XA(19)   <= '0';      XA(18) <= '0';
        SRAM_OE  <= CLK02;    FLASH_OE <= '1';
        SRAM_CE  <= CLK02;    FLASH_CE <= '1';
      end if;

    when WROMWrite =>
      XD(15) <= '0';        XD(14) <= '0';
      XD(13) <= '0';        XD(12) <= '0';
      XD(11) <= '0';        XD(10) <= '0';
      
      XD(09) <= DSR(09);    XD(08) <= DSR(08);
      XD(07) <= DSR(07);    XD(06) <= DSR(06);
      XD(05) <= DSR(05);    XD(04) <= DSR(04);
      XD(03) <= DSR(03);    XD(02) <= DSR(02);
      XD(01) <= DSR(01);    XD(00) <= DSR(00);

      XA(19) <= '0';        XA(18) <= '0';
      XA(17) <= SR(05);     XA(16) <= SR(04);
      XA(15) <= SR(03);     XA(14) <= SR(02);
      XA(13) <= SR(01);     XA(12) <= SR(00);
      XA(11) <= DSR(23);    XA(10) <= DSR(22);
      XA(09) <= DSR(21);    XA(08) <= DSR(20);
      XA(07) <= DSR(19);    XA(06) <= DSR(18);
      XA(05) <= DSR(17);    XA(04) <= DSR(16);
      XA(03) <= DSR(15);    XA(02) <= DSR(14);
      XA(01) <= DSR(13);    XA(00) <= DSR(12);

      SRAM_OE <= '1';       FLASH_OE <= '1';
      SRAM_CE <= CLK02;     FLASH_CE <= '1';
      SRAM_WE <= CLK02;     FLASH_WE <= '1';

    when MLDL_disable =>
       -- keep feeding the shift registers to the outputs for prototyping
      XA(19) <= DSR(31);    XA(18) <= DSR(30);
      XA(17) <= DSR(29);    XA(16) <= DSR(28);
      XA(15) <= DSR(27);    XA(14) <= DSR(26);
      XA(13) <= DSR(25);    XA(12) <= DSR(24);
      XA(11) <= DSR(23);    XA(10) <= DSR(22);
      XA(09) <= DSR(21);    XA(08) <= DSR(20);
      XA(07) <= DSR(19);    XA(06) <= DSR(18);
      XA(05) <= DSR(17);    XA(04) <= DSR(16);
      XA(03) <= DSR(15);    XA(02) <= DSR(14);
      XA(01) <= DSR(13);    XA(00) <= DSR(12);

      if (X_STROBE = '0') then
         -- Address lines are already driven from DSR[30..12]
        if DSR(11) = '0' then                       -- write command
          XD(15) <= DSR(09);    XD(14) <= DSR(08);
          XD(13) <= DSR(07);    XD(12) <= DSR(06);
          XD(11) <= DSR(05);    XD(10) <= DSR(04);
                
          XD(09) <= DSR(03);    XD(08) <= DSR(02);
          XD(07) <= DSR(01);    XD(06) <= DSR(00);
          XD(05) <= XSR(05);    XD(04) <= XSR(04);
          XD(03) <= XSR(03);    XD(02) <= XSR(02);
          XD(01) <= XSR(01);    XD(00) <= XSR(00);
          SRAM_OE <= '1';       FLASH_OE <= '1';

          if DSR(10) = '0' then                     -- We are talking to FLASH
            SRAM_CE <= '1';         FLASH_CE <= '0';
            SRAM_WE <= '1';         FLASH_WE <= not X_CLK;
           else                                     -- SRAM is addressed
            SRAM_CE <= '0';         FLASH_CE <= '1';
            SRAM_WE  <= not X_CLK;  FLASH_WE <= '1';
          end if;
        else    -- DSR(11) = '1'                    -- read command
          XD <= (others => 'Z');
          SRAM_WE <= '1';       FLASH_WE <= '1';
          if DSR(10) = '0' then                     -- We are talking to FLASH
            SRAM_CE <= '1';         FLASH_CE <= '0';
            SRAM_OE <= '1';         FLASH_OE <= '0';
          else                                      -- SRAM is addressed
            SRAM_CE <= '0';         FLASH_CE <= '1';
            SRAM_OE <= '0';         FLASH_OE <= '1';
          end if;
        end if;
      else                                          -- STROBE = '1'
        SRAM_CE <= '1';       FLASH_CE <= '1';
        SRAM_OE <= '1';       FLASH_OE <= '1';
        SRAM_WE <= '1';       FLASH_WE <= '1';
        XD <= (others => 'Z');
      end if;

    when IORead =>
      SRAM_WE <= '1';       FLASH_WE <= '1';
      XD <= (others => 'Z');
      if ISR(11) = '0' and ISR(10) = '1' then       -- Read Memory
        if ISR(09) = '0' then                       -- Memory
          XA(19) <= DSR(31);    XA(18) <= DSR(30);
          XA(17) <= DSR(29);    XA(16) <= DSR(28);
          XA(15) <= IOR(07);    XA(14) <= IOR(06);
          XA(13) <= IOR(05);    XA(12) <= IOR(04);
          XA(11) <= IOR(03);    XA(10) <= IOR(02);
          XA(09) <= IOR(01);    XA(08) <= IOR(00);
        else                                        -- SR access
          XA(19) <= '1';        XA(18) <= '1';
          XA(17) <= '1';        XA(16) <= '1';
          XA(15) <= '1';        XA(14) <= '1';
          XA(13) <= '1';        XA(12) <= '1';
          XA(11) <= '1';        XA(10) <= '1';
          XA(09) <= '1';        XA(08) <= '1';
        end if;
          XA(07) <= DSR(19);    XA(06) <= DSR(18);
          XA(05) <= DSR(17);    XA(04) <= DSR(16);
          XA(03) <= DSR(15);    XA(02) <= DSR(14);
          XA(01) <= DSR(13);    XA(00) <= DSR(12);

        if DSR(10) = '0' then                       -- FLASH is addressed
          SRAM_CE <= '1';       FLASH_CE <= CLK02;
          SRAM_OE <= '1';       FLASH_OE <= CLK02;
        else                                        -- SRAM is addressed
          SRAM_OE  <= CLK02;    FLASH_OE <= '1';
          SRAM_CE  <= CLK02;    FLASH_CE <= '1';
        end if;
      else                            -- Real I/O or outside window, do nothing
        SRAM_CE <= '1';       FLASH_CE <= '1';
        SRAM_OE <= '1';       FLASH_OE <= '1';
        XD <= (others => 'Z');
        XA <= (others => '0');
      end if;

    when IOWrite =>
      SRAM_OE <= '1';      FLASH_OE <= '1';
      if DSR(23) = '0' and DSR(22) = '1' then       -- Write Memory
        if DSR(21) = '0' then                       -- Memory
          XA(19) <= DSR(31);    XA(18) <= DSR(30);
          XA(17) <= DSR(29);    XA(16) <= DSR(28);
          XA(15) <= IOR(07);    XA(14) <= IOR(06);
          XA(13) <= IOR(05);    XA(12) <= IOR(04);
          XA(11) <= IOR(03);    XA(10) <= IOR(02);
          XA(09) <= IOR(01);    XA(08) <= IOR(00);
        else                                        -- SR access
          XA(19) <= '1';        XA(18) <= '1';
          XA(17) <= '1';        XA(16) <= '1';
          XA(15) <= '1';        XA(14) <= '1';
          XA(13) <= '1';        XA(12) <= '1';
          XA(11) <= '1';        XA(10) <= '1';
          XA(09) <= '1';        XA(08) <= '1';
        end if;
        XA(07) <= DSR(19);    XA(06) <= DSR(18);
        XA(05) <= DSR(17);    XA(04) <= DSR(16);
        XA(03) <= DSR(15);    XA(02) <= DSR(14);
        XA(01) <= DSR(13);    XA(00) <= DSR(12);

        XD(15) <= '0';        XD(14) <= '0';
        XD(13) <= '0';        XD(12) <= '0';
        XD(11) <= '0';        XD(10) <= '0';

        XD(09) <= DSR(09);    XD(08) <= DSR(08);
        XD(07) <= DSR(07);    XD(06) <= DSR(06);
        XD(05) <= DSR(05);    XD(04) <= DSR(04);
        XD(03) <= DSR(03);    XD(02) <= DSR(02);
        XD(01) <= DSR(01);    XD(00) <= DSR(00);

        SRAM_CE <= CLK02;     FLASH_CE <= '1';
        SRAM_WE <= CLK02;     FLASH_WE <= '1';
      else                            -- Real I/O or outside window, do nothing
        SRAM_CE <= '1';       FLASH_CE <= '1';
        SRAM_WE <= '1';       FLASH_WE <= '1';
        XD <= (others => 'Z');
        XA <= (others => '0');
      end if;

    when others =>
      SRAM_CE <= '1';     FLASH_CE <= '1';
      SRAM_OE <= '1';     FLASH_OE <= '1';
      SRAM_WE <= '1';     FLASH_WE <= '1';
      XD <= (others => 'Z');
      XA <= (others => '0');

  end case;

end process;


-------------------------------------------------------------------------------
--             OTHER SIGNALS DRIVER                                         --
-------------------------------------------------------------------------------

   -- unused HP41 outputs, all are disabled!
  DATA_out <= '0';
  DATA_OE  <= 'Z';
  FI_out   <= '0';
  FI_OE    <= 'Z';
  SYNC_out <= '0';
  SYNC_OE  <= 'Z';

  MEM_RESET <= '1';            -- no reset, implement later????
  SRAM_CE1  <= SRAM_CE;
  SRAM_CE2  <= not SRAM_CE;


end Behavioral;
